% SWI-Prolog program

% Reading/writing ODE systems in XPP format (as generated by Biocham)
% and Biocham models


% F. Fages, Inria, March 2020

:- module(
  minibiocham,
  [
    bc2pl/1,
    print_crn/0,

    pl2xpp/1,
    print_xpp/0,

    reaction/2,
    species/1,
    parameter/2,
    initial/2,

    new_ode/0,
    ode_init/2,
    ode_par/2,
    ode_func/2,
    ode_dxdt/2,

    % Biocham syntax
    op(1100, xfx, for),
    op(1050, xfx, =>),

    % XPP syntax
    op(1010, fx, init),
    op(1010, fx, par),
    op(1010, fx, num),
    op(1010, fx, func)
  ]
).


%%%%%
% CRN representation in Prolog
%%%%%
:- dynamic(reaction/2). % databse of reactions given with kinetics and change vector
:- dynamic(species/1).  % database of molecular species
:- dynamic(initial/2).  % database of initial states
:- dynamic(parameter/2).% database of parameters


new_crn:-
  retractall(reaction(_,_)),
  retractall(species(_)),
  retractall(parameter(_,_)),
  retractall(initial(_, _)),
  new_ode.


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Parsing Biocham reactions
%%%%%%%%%%%%%%%%%%%%%%%%%%%

bc2pl(InputFile):-
  new_crn,
  open(InputFile, read, Stream),
  read_crn(Stream),
  close(Stream).


read_crn(Stream):-
  repeat,     % iteration until stopped by cut predicate "!"
  read_term(Stream, Term, [var_prefix(true), variable_names(VarNameList)]), 
  maplist(name_variable, VarNameList), % instanciates the tokens read as Prolog variables to their names
  debug(crn, "read term: ~k", [Term]),
  (
    Term=end_of_file
  ->
    ! 
  ;
    (
      Term=(L => R)
    ->
      add_reaction('MA'(1), L, R)
    ;
      Term=(K for L => R)
    ->
      add_reaction(K, L, R)
    ;
      Term=present(A)
    ->
      assertz(initial(A,1))
    ;
      Term=present(A,V)
    ->
      assertz(initial(A,V))
    ;
      Term =.. [parameter | L] % term decomposition in a list [Head | Arguments]
    ->
      forall(member(P=V, L), assertz(parameter(P,V)))
    ;
      write('skiping ')
    ),
    writeln(Term),
    fail
  ).

name_variable(Name=Var):-
  (atom_prefix(Name, '_') % particular case of the character _ which is read in Prolog as a new variable with a number added
->
  Var='_'
;
  Var=Name % instanciates the Prolog variable to its name to become a ground term
).


% extracts the change vector from the stoichiometry of reactants and products
add_reaction(Kinetics, Reactants, Products):-
  debug(crn, "add_reaction(~w, ~w, ~w)", [Kinetics, Reactants, Products]),
  add_reactants(Reactants, List),
  debug(crn, "add_reactants -> ~w", [List]),
  add_products(Products, List, Changes),
  debug(crn, "add_products -> ~w", [Changes]),
  (
    Kinetics='MA'(K)
  ->
    mass_action_kinetics(List, K, Rate)
  ;
    Rate=Kinetics
  ),
  debug(crn, "kinetics: ~w", [Rate]),
  assertz(reaction(Rate, Changes)).

% species with negative stoichiometric coefficients for reactants
add_reactants(Solution, List):-
  (
    Solution='_'
  ->
    List=[]
  ;
    Solution=S1+S2
  ->
    add_reactants(S1, L1),
    add_reactants(S2, L2),
    append(L2, L1, List)
  ;
    Solution=K*Molecule
  ->
    Km is -K,
    List=[(Molecule, Km)],
    add_species(Molecule)
  ;
    List=[(Solution, -1)],
    add_species(Solution)
  ).


add_products(Solution, List, Changes):-
  (
    Solution='_'
  ->
    Changes=List
  ;
    Solution=S1+S2
  ->
    add_products(S1, List, L),
    add_products(S2, L, Changes)
  ;
    Solution=K*Molecule
  ->
    update_changes(List, Molecule, K, Changes),
    add_species(Molecule)
  ;
    update_changes(List, Solution, 1, Changes),
    add_species(Solution)
  ).


mass_action_kinetics(List, Expression, Rate) :-
  mass_action_kinetics(List, Product),
  (
    Expression == 1
  ->
    Rate = Product
  ;
    Rate = Expression*Product
  ).

mass_action_kinetics([], 1).
mass_action_kinetics([(Molecule, Coefficient)], R):-
  C is -Coefficient, % reactants in the list have a negative stoichiometric coefficient
  (
    C=1
  ->
    R=Molecule
  ;
    R=Molecule^C
  ).
mass_action_kinetics([(Molecule, Coefficient) | List], Product * R):-
  List = [_ | _],
  C is -Coefficient, 
  (
    C=1
  ->
    R=Molecule
  ;
    R=Molecule^C
  ),
  mass_action_kinetics(List, Product).


update_changes([], Molecule, K, [(Molecule, K)]).
update_changes([(M, V) | List], Molecule, K, NewChanges):-
  (
    M=Molecule
  ->
    R is V+K,
    (
      R = 0
    ->
      NewChanges = List
    ;
      NewChanges = [(M, R) | List]
    )
  ;
    NewChanges = [(M, V) | Changes],
    update_changes(List, Molecule, K, Changes)
  ).


add_species(Molecule):-
  (
    species(Molecule)
  ->
    true
  ;
    assertz(species(Molecule))
  ).


%%%%%%%%%%%%%%%%%%
% pretty-print CRN
%%%%%%%%%%%%%%%%%%

print_crn:-
  listing(species/1),
  listing(reaction/2),
  listing(initial/2),
  listing(parameter/2).



%%%%%%%%%%%%
% ODE system representation in Prolog
%%%%%%%%%%%%

% predicates with dynamic clauses (assert/retract) used as global memory
:- dynamic(ode_init/2).
:- dynamic(ode_par/2).
:- dynamic(ode_func/2).
:- dynamic(ode_dxdt/2).


new_ode:-
  retractall(ode_init(_,_)),  % erase dynamic global memory facts
  retractall(ode_par(_,_)),
  retractall(ode_func(_,_)),
  retractall(ode_dxdt(_,_)).


%%%%%%%%%%%%%%
% Pretty print ODE system in XPP format
%%%%%%%%%%%%%%



% pretty-print all global memory facts about ODE system
print_xpp :-
  forall(ode_init(X, E), format('init ~w = ~w\n', [X, E])),
  forall(ode_par(K, V), format('par ~w = ~w\n', [K, V])),
  forall(ode_func(F, D), format('~w = ~w\n', [F, D])),
  forall(ode_dxdt(X, E), format('d~w/dt = ~w\n', [X, E])),
  format('done\n').


%%%%%%%%%%%%%%%%%%%%%
% Writing to XPP file
%%%%%%%%%%%%%%%%%%%%%

pl2xpp(OutputFile) :-
  open(OutputFile, write, Stream),
  set_output(Stream),
  print_xpp,
  close(Stream).
