#include <stdlib.h>
#include <cstddef>
#include <random>
#include <chrono>
#include <ppl.hh>
#include "ppl_validity_domain.hh"
#include "gsl_solver.hh"
#include "cmaes_interface.h"
#include "satisfaction_degree.hh"
#include "ode.inc"
#include "check.inc"
#include "search_parameters.inc"
#include "domain_distance.inc"
#include "filter.inc"

#include <mpi.h>

static double
satisfaction_degree(gsl_solver *solver, gsl_solver_config *config, double values[DIMENSION])
{
    gsl_solver_state state;
    int j;
    double *p = solver->p;
    Table table;
    double max_distance;
    Domain domain;

    init_gsl_solver_state(&state, solver);
    // reset step size and evolution, forces again initial step size because
    // otherwise seems useless
    gsl_odeiv2_driver_reset_hstart(solver->d, config->initial_step_size);
    for (j = 0; j < DIMENSION; j++) {
        const struct search_parameter *param = &search_parameters[j];
        double value = values[j];
        p[param->index] = value;
        /* std::cerr << "param " << j << " = "  << value << std::endl; */
    }

    // initial values might depend on parameters
    initial_values(state.x, p);

    gsl_solver_iterate(&state, &table);
    free_gsl_solver_state(&state);

    PPL::set_rounding_for_PPL();
    domain = (*compute_condition_domain[0])(table);
    /* using PPL::IO_Operators::operator<<; */
    /* std::cerr << domain << "." << std::endl; */
    /* std::cerr << j << " rows" << std::endl; */
    // Ignore negative distances for robustness computation
    max_distance = std::max(0.0, domain_distance(domain, 0));
    /* std::cerr << "distance " << max_distance << std::endl; */
    PPL::restore_pre_PPL_rounding();

    return 1/(1 + max_distance);
}


int
main(int argc, char *argv[])
{
    double population[DIMENSION];
    std::default_random_engine generator;
    if (SEED > 0) {
        generator.seed(SEED);
    } else {
        generator.seed(std::chrono::system_clock::now().time_since_epoch().count());
    }
    std::normal_distribution<double> distribution[DIMENSION];
    int i, j;
    // https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
    double robustness_mean = 0, new_value;
    // double robustness_ssd = 0, delta, delta2, relative_robustness_error;
    gsl_solver_config config;
    gsl_solver solver;
    int loc_samples;

    MPI_Init(NULL, NULL);
    // Get the number of processes
    int world_size;
    MPI_Comm_size(MPI_COMM_WORLD, &world_size);
    // Get the rank of the process
    int world_rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    // fprintf(stderr, "proc %d of %d\n", world_rank, world_size);

    // Different seed on each processor
    generator.seed(SEED*(world_rank + 1));
    init_gsl_solver_config(&config);
    init_gsl_solver(&solver, &config);
    for (i = 0; i < DIMENSION; i++) {
        const struct search_parameter *param = &search_parameters[i];
        // fprintf(stderr, "param %d init %g\n", i, param->init);
        distribution[i] = std::normal_distribution<double>(param->init, COEFF_VAR*param->init);
    }
    if (world_rank == 0) {
        loc_samples = (int) SAMPLES / world_size + SAMPLES % world_size;
    } else {
        loc_samples = (int) SAMPLES / world_size;
    }
    // fprintf(stderr, "proc %d handling %d samples\n", world_rank, loc_samples);
    new_value = 0;
    for (j = 0; j < loc_samples; ++j) {
        for (i = 0; i < DIMENSION; i++) {
            population[i] = distribution[i](generator);
        }
        new_value += satisfaction_degree(&solver, &config, population);
        // NOTE removed in favor of MPI parallel code
        // delta = new_value - robustness_mean;
        // robustness_mean += delta/(j+1);
        // delta2 = new_value - robustness_mean;
        // robustness_ssd += delta*delta2;
        // https://en.wikipedia.org/wiki/Monte_Carlo_integration
        // https://en.wikipedia.org/wiki/Variance#Population_variance_and_sample_variance
        // relative_robustness_error = sqrt(robustness_ssd)/((j+1)*robustness_mean);
        // fprintf(stderr, "%d: %g %g\n", j, robustness_mean, relative_robustness_error);
        // if (j > SAMPLES*RELATIVE_ROBUSTNESS_ERROR &&
        //         relative_robustness_error < RELATIVE_ROBUSTNESS_ERROR) {
        //     break;
        // }
    }

    // fprintf(stderr, "proc %d total robustness %f\n", world_rank, new_value);
    MPI_Reduce(&new_value, &robustness_mean, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    if (world_rank == 0) {
        // fprintf(stderr, "proc %d sum of robustness %f\n", world_rank, robustness_mean);
        robustness_mean = robustness_mean / SAMPLES;
        printf("%f\n", robustness_mean);
    }
    free_gsl_solver(&solver);

    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Finalize();

    return EXIT_SUCCESS;
}
