#include <stdlib.h>
#include <cstddef>
#include <ppl.hh>
#include "ppl_validity_domain.hh"
#include "gsl_solver.hh"
#include "cmaes_interface.h"
#include "satisfaction_degree.hh"
#include "ode.inc"
#include "check.inc"
#include "search_parameters.inc"
#include "domain_distance.inc"
#include "filter.inc"

#include <mpi.h>

static double
fitfun(gsl_solver *solver, gsl_solver_config *config, double values[DIMENSION])
{
    gsl_solver_state state;
    int i, j;
    double *p = solver->p;
    Table table;
    double result = 0;
    double max_distance = -INFINITY;
    Domain domain;

    for (i = 0; i < CONDITIONS; i++) {
        init_gsl_solver_state(&state, solver);
        // reset step size and evolution, forces again initial step size because
        // otherwise seems useless
        gsl_odeiv2_driver_reset_hstart(solver->d, config->initial_step_size);
        for (j = 0; j < DIMENSION; j++) {
            const struct search_parameter *param = &search_parameters[j];
            double value = values[j];
            if (i == 0) { // first condition
                if (value < param->min) {
                    result += param->min - value;
                }
                else if (value > param->max) {
                    result += value - param->max;
                }
            }
            if (LOGNORMAL) {
                p[param->index] = exp(value);
            } else {
                p[param->index] = value;
            }
            /* std::cerr << "param " << j << " = "  << value << std::endl; */
        }
        // first condition, out-of-bounds penalty
        if (i == 0 && result > 0) {
            result += 10000;
            // if violation, simply return with no simulation
            return result;
        }
        (*setup_condition[i])(solver);

        // initial values might depend on parameters
        initial_values(state.x, p);

        gsl_solver_iterate(&state, &table);
        free_gsl_solver_state(&state);

        PPL::set_rounding_for_PPL();
        domain = (*compute_condition_domain[i])(table);
        /* using PPL::IO_Operators::operator<<; */
        /* std::cerr << domain << "." << std::endl; */
        /* std::cerr << j << " rows" << std::endl; */
        max_distance = std::max(max_distance, domain_distance(domain, i));
        /* std::cerr << "distance " << max_distance << std::endl; */
        PPL::restore_pre_PPL_rounding();
    }

    result += max_distance;
    return result;
}


int
main(int argc, char *argv[])
{
    cmaes_t evo;
    double *arFunvals, *const*pop, *xfinal;
    double xstart[DIMENSION], stddev[DIMENSION];
    int i;
    /* int j; */
    int lambda;
    gsl_solver_config config;
    gsl_solver solver;
    double best_vd = 1;

    MPI_Init(NULL, NULL);
    // Get the number of processes
    int world_size;
    MPI_Comm_size(MPI_COMM_WORLD, &world_size);
    // Get the rank of the process
    int world_rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    // fprintf(stderr, "proc %d of %d\n", world_rank, world_size);
    MPI_Request req;

    init_gsl_solver_config(&config);
    init_gsl_solver(&solver, &config);
    if (world_rank != 0) {
        double data[DIMENSION];
        double fit;
        MPI_Status status;
        int amount;
        while(true) {
            // fprintf(stderr, "proc %d ready\n", world_rank);
            MPI_Recv(&data, DIMENSION, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
            MPI_Get_count(&status, MPI_DOUBLE, &amount);
            if (amount == DIMENSION) {
                // fprintf(stderr, "proc %d received\n", world_rank);
                fit = fitfun(&solver, &config, data);
                // fprintf(stderr, "proc %d sending %f\n", world_rank, fit);
                MPI_Isend(&fit, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &req);
                MPI_Request_free(&req);
            } else {
                // fprintf(stderr, "proc %d terminating\n", world_rank);
                break;
            }
        }
    } else {
        for (i = 0; i < DIMENSION; i++) {
            const struct search_parameter *param = &search_parameters[i];
            double min = param->min;
            double max = param->max;
            xstart[i] = param->init;
            stddev[i] = (max - min) / 2;
        }
        arFunvals = cmaes_init(
            &evo, DIMENSION, xstart, stddev, SEED, 0, "cmaes_initials.par");
        cmaes_ReadSignals(&evo, "cmaes_signals.par");
        evo.sp.stStopFitness.flg = 1;
        evo.sp.stStopFitness.val = STOPFIT;
        // somehow still have to test for best_vd, not sure why...
        lambda = cmaes_Get(&evo, "lambda");
        // fprintf(stderr, "population size %d spread on %d procs\n", lambda, world_size);

        while(!cmaes_TestForTermination(&evo) && best_vd >= STOPFIT) {
            pop = cmaes_SamplePopulation(&evo);
            // Send all in a non-blocking way
            for (i = 0; i < lambda; ++i) {
                if (i % world_size != 0) {
                    // fprintf(stderr, "proc %d sending to %d\n", world_rank, i % world_size);
                    MPI_Isend(pop[i], DIMENSION, MPI_DOUBLE, i % world_size, 0, MPI_COMM_WORLD, &req);
                    MPI_Request_free(&req);
                }
            }
            // Do all locally
            for (i=0; i < lambda; ++i) {
                if (i % world_size == 0) {
                    arFunvals[i] = fitfun(&solver, &config, pop[i]);
                    // fprintf(stderr, "proc %d computing locally\n", world_rank, arFunvals[i]);
                }
            }
            // Wait for all results
            for (i=0; i < lambda; ++i) {
                if (i% world_size != 0) {
                    // fprintf(stderr, "proc %d waiting for %d\n", world_rank, i % world_size);
                    MPI_Recv(&arFunvals[i], 1, MPI_DOUBLE, i % world_size, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                    // fprintf(stderr, "proc %d received %f from %d\n", world_rank, arFunvals[i], i % world_size);
                }
                /* fprintf(stderr, "distance: %f\n", arFunvals[i]); */
                /* for (j = 0; j < DIMENSION; ++j) { */
                /*     fprintf(stderr, "   %.17g\n", pop[i][j]); */
                /* } */
            }
            // fprintf(stderr, "proc %d finished one loop\n", world_rank);
            cmaes_UpdateDistribution(&evo, arFunvals);
            best_vd = cmaes_Get(&evo, "fbestever");
            /* fprintf(stderr, "Best satisfaction degree: %f\n", */
            /*         1 / (1 + best_vd)); */
            /* xfinal = cmaes_GetNew(&evo, "xbestever"); */
            /* for (i = 0; i < DIMENSION; i++) { */
            /*     fprintf(stderr, "   %f\n", xfinal[i]); */
            /* } */
            cmaes_ReadSignals(&evo, "cmaes_signals.par");
            fflush(stdout);
        }
        for (i=1; i < world_size; i++) {
            // send end message
            MPI_Send(&best_vd, 0, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
        }
        // fprintf(stderr, "proc %d done\n", world_rank);
        cmaes_WriteToFile(&evo, "all", "allcmaes.dat");
        if (cmaes_TestForTermination(&evo)) {
            char reason[3024];
            strncpy(reason, cmaes_TestForTermination(&evo), 3024);
            strtok(reason, "\n");
            printf("%s\n", reason);
        } else {
            printf("Biocham stop with best fitness %g < stop fitness %g\n",
                   best_vd, STOPFIT);
        }
        xfinal = cmaes_GetNew(&evo, "xbestever");
        for (i = 0; i < DIMENSION; i++) {
            /* fprintf(stderr, "%.17g\n", xfinal[i]); */
            if (LOGNORMAL) {
                printf("%.17g\n", exp(xfinal[i]));
            }
            else {
                printf("%.17g\n", xfinal[i]);
            }
        }
        printf("%f\n", 1 / (1 + best_vd));
        cmaes_exit(&evo);

        free(xfinal);
    }
    free_gsl_solver(&solver);

    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Finalize();

    return EXIT_SUCCESS;
}
