:- module(matrix,
  [
    constant_list/3,
    constant_matrix/4,
    replace/4,
    get_element/3,
    replace_element/3,
    init_matrix/1,
    write_matrix/1,
    ludcmp_matrix/2,
    ludcmp_matrix/0,
    matrix_sum/4,
    matrix_prod/3,
    scalar_prod/3,
    matrix_apply/3,
    matrix_sum/3,
    create_diag_matrix/3,
    sub_matrix/1,
    transpose_matrix/2,
    lusubst/2,
    solve_linear_problem/3
  ] 
).

:- use_module(util).

/* Module Matrix
*
* Several operations are done on a "global variable" of SWI-prolog (using nb_set/getval)
* but this is due to historical version of an efficient implementation in GNU-prolog and
* may not be necessary nor useful now.
*/


%! constant_list(+Length, +Element, -List)
%
% create a list with every element set up to Element

constant_list(0,_,[]) :-
  !. 

constant_list(N1,X,[X|L]) :- 
  N1 > 0, N is N1 - 1,    
  constant_list(N,X,L).


%! constant_matrix(+N_row, +N_column, +Element, -List)
%
% create a matrix with every element set up to Element

constant_matrix(RowSize,ColumnSize,Value,Matrix) :-
  constant_list(ColumnSize,Value,Row), % The row size is the number of column
  constant_list(RowSize,Row,Matrix). 


%! replace(+List, +Index, +NewElement, -NewList)
%
% Allows to replace the i-th element of the list, where the 1st element is index by 1

replace([_|Tail], 1, NewElement, [NewElement|Tail]) :- 
  !.
   
replace([Element|Tail], I, NewElement, [Element|Tail2]):-
  I > 1, I1 is I-1,
  replace(Tail, I1, NewElement, Tail2).   


%! init_matrix(+Matrix)
%
% Creates the global variable which will stores the matrix on which we are doing operations 

init_matrix([Head|Tail]):-
  length([Head|Tail],RowSize),
  length(Head,ColumnSize),
  nb_setval(stored_matrix,[Head|Tail]),
  nb_setval(formal_m,RowSize),
  nb_setval(formal_n,ColumnSize).


%! get_element(+RowIndex, +ColumnIndex, -Value)
%
% get the (i,j)-th element of the stored_matrix

get_element(RowIndex,ColumnIndex,Value) :- 
  nb_getval(stored_matrix,Matrix),
  nth1(RowIndex,Matrix,Row),
  nth1(ColumnIndex,Row,Value).


%! replace_element(+RowIndex, +ColumnIndex, +Value) :- 
%
% set the (i,j)-th element of the stored matrix to Value

replace_element(RowIndex,ColumnIndex,Value) :- 
  nb_getval(stored_matrix,Matrix),
  nth1(RowIndex,Matrix,Row),
  replace(Row,ColumnIndex,Value,NewRow),
  replace(Matrix,RowIndex,NewRow,NewMatrix),
  nb_setval(stored_matrix,NewMatrix).


/* Some tools to write matrix in a pretty way, for debugging for examples*/
%! write_matrix(+Matrix)

write_matrix([]).

write_matrix([H|T]):-
  write_tab(H),
  write_matrix(T).


%! write_tab(+List)

write_tab([]) :- 
  nl.

write_tab([H|T]):-
  write(H),write('\t'),
  write_tab(T).

%! ludcmp_matrix(+Matrix, -TransformedMatrix)
%
% Perform an LU decomposition, that is find two matrices L and U such that Matrix = L.U,
% L and U begin respectively lower and upper diagonal.
% By convention, L has only 1 on its diagonal so that we gather them in a single matrix M
% such that M[i,j] = L[i,j] if i<j and M[i,j] = U[i,j] if i>=j.  M is the transformed
% matrix and is stored in the global variable at the end of the call.

ludcmp_matrix(A,B):-
  init_matrix(A),
  ludcmp_matrix,
  nb_getval(stored_matrix,B).


ludcmp_matrix :-
   % Value if pivot element is zero (singular)
   TINY=1.0E-100,
   nb_getval(formal_m,M),
   nb_getval(formal_n,N),
   (
      M \= N
   ->
      throw(error('Matrix must be square for LU decomposition')) 
   ;
      true
   ),
   constant_list(N,0,X), nb_setval(formal_index,X),
   constant_list(N,0,Y), nb_setval(formal_scaling,Y),
   % fill implicit scaling info, formal_scaling contains 1/RowMax where RowMax is the max(abs(RowElement)) for each Row
   forall(
         between(1,N,I),
         (nb_setval(formal_tmp,0),
         forall(
               between(1,N,J),
               (get_element(I,J,T2),
               TMP is abs(T2),
               nb_getval(formal_tmp,T1),
               (
                  TMP > T1
               ->
                  nb_setval(formal_tmp,TMP)
               ;
                  true
               )
               )
         ),
         nb_getval(formal_tmp,TMP),
         %format("formal_tmp à ~w est ~w",[I,TMP]), nl ,
         (
            TMP =:= 0
         ->
            throw(error(singular_matrix)) 
         ;
            true
         ),
         T3 is 1/TMP,
         nb_getval(formal_scaling,Array), replace(Array,I,T3,NewArray), nb_setval(formal_scaling,NewArray)
         )
   )
   ,  
   forall(
         between(1,N,J),
         (nb_setval(formal_tmp,0),
         forall(
               between(1,N,I),
               (%format("computing (~w,~w)~n",[I,J]),
               (
               % I=<J
                  I<J
               ->
                  matrix_sum(I,I,J,S),
                  replace_element(I,J,S)
               ;
                  matrix_sum(J,I,J,S),
                  replace_element(I,J,S),
                  nb_getval(formal_scaling,Scaling), nth1(I,Scaling,VV),
                  TMP is abs(S)*VV,
                  nb_getval(formal_tmp,T1),
                  (
                     TMP > T1
                  ->
                     nb_setval(formal_tmp,TMP),
                     nb_setval(formal_imax,I)
                  ;
                     true
                  )
               )
               )     
         ),
         nb_getval(formal_imax,IMAX),
         %format("for ~w max row is ~w~n",[J,IMAX]),
         (
            J \= IMAX
         ->
            % we need to interchange rows
            forall(
                  between(1,N,II),
                  (get_element(IMAX,II,DUM),
                  get_element(J,II,TMP),
                  replace_element(IMAX,II,TMP),
                  replace_element(J,II,DUM)             
                  )
            ),
            nb_getval(formal_scaling,Scaling),nth1(J,Scaling,VV),
            replace(Scaling,IMAX,VV,NewScaling), nb_setval(formal_scaling,NewScaling)
         ;
            true
         ),
         nb_getval(formal_index,Index), replace(Index,J,IMAX,NewIndex),nb_setval(formal_index,NewIndex),
         get_element(J,J,AJJ),
         (
            AJJ =:= 0
         ->
            % don't keep a zero here, we will divide by it!
            AAJJ = TINY,
            replace_element(J,J,AAJJ)
         ;
            AAJJ = AJJ
         ), 
         (
            J < N
         ->
            JJ is J+1,
            forall(
                  between(JJ,N,II),
                  (get_element(II,J,AIJ),
                  AAIJ is AIJ/AAJJ,
                  replace_element(II,J,AAIJ)
                  )
            )
         ;
            true
         )     
         )
   ).

%! matrix_sum(+N, +I, +J, -Sum)
%
% if M denotes the stored matrix, return the sum for K between 1 and N of M(I,K)M(K,J)

matrix_sum(SumSize,RowIndex,ColumnIndex,Sum):-
  get_element(RowIndex,ColumnIndex,A),
  matrix_sum_rec(1,SumSize,RowIndex,ColumnIndex,A,Sum).


matrix_sum_rec(SumSize,SumSize,_,_,Sum,Sum):- 
  !.

matrix_sum_rec(SumIndex,SumSize,RowIndex,ColumnIndex,PartialSum,Sum):-
  get_element(RowIndex,SumIndex,IK),
  get_element(SumIndex,ColumnIndex,KJ),
  PartialSum2 is PartialSum - IK*KJ,
  NewSumIndex is SumIndex+1,
  matrix_sum_rec(NewSumIndex,SumSize,RowIndex,ColumnIndex,PartialSum2,Sum).


 %! matrix_prod(+A, +B, -P)
 %
 % P = A.B

matrix_prod(MatrixA,MatrixB,Result):-
  transpose_matrix(MatrixB,TransposedMatrixB),
  matrix_prod_rec(MatrixA,TransposedMatrixB,Result).

matrix_prod_rec([],_,[]).
matrix_prod_rec([RowA|MatrixA],MatrixB,[RowResult|MatrixResult]):-
  matrix_prod_sub(RowA,MatrixB,RowResult),
  matrix_prod_rec(MatrixA,MatrixB,MatrixResult).

matrix_prod_sub(_,[],[]).
matrix_prod_sub(Row,[RowB|MatrixB],[ElementResult|RowResult]):-
  scalar_prod(Row,RowB,ElementResult),
  matrix_prod_sub(Row,MatrixB,RowResult).


%! scalar_prod(+X, +Y, -Product)
%
% Comput the scalar product of X and Y

scalar_prod(VecX, VecY, SP) :-
  scalar_prod(VecX, VecY, 0, SP).

scalar_prod([], [], SP, SP).

scalar_prod([X|TailX], [Y|TailY], Res_tempo, Acc):-
  Res is Res_tempo + X*Y,
  scalar_prod(TailX, TailY, Res, Acc).


%! matrix_apply(+M, +V, -Res)
%
% compute the dot product of M.V

matrix_apply([],_,[]).

matrix_apply([Row|Matrix],Vector,[ResultElement|ResultVector]):-
  scalar_prod(Row,Vector,ResultElement),
  matrix_apply(Matrix,Vector,ResultVector).
   

%! matrix_sum(+A, +B, -Sum)
%
% Compute the element sum of two matrices

matrix_sum([],[],[]).

matrix_sum([RowA|MatrixA],[RowB|MatrixB],[SumRow|SumMatrix]):-
  add_list(RowA,RowB,SumRow),
  matrix_sum(MatrixA,MatrixB,SumMatrix).


%! create_diag_matrix(+Size, +Value, -Diag_matrix)
%
% Store a diagonal matrix in the global variable and return it

create_diag_matrix(Size,Value,DiagMatrix):-
  constant_matrix(Size,Size,0,Matrix),
  nb_setval(stored_matrix,Matrix),
  nb_setval(formal_n,Size),
  nb_setval(formal_m,Size),
  forall(
    between(1,Size,RowIndex),
    replace_element(RowIndex,RowIndex,Value)
  ),
  nb_getval(stored_matrix,DiagMatrix).


% sub_matrix(+Matrix)
%
% Substract Matrix from the global matrix

sub_matrix([Row|Matrix]):-
  sub_matrix([Row|Matrix],1).

sub_matrix([],_).
sub_matrix([Row|Matrix],RowIndex):-
  sub_matrix_rec(Row,RowIndex,1),
  NewRowIndex is RowIndex+1,
  sub_matrix(Matrix,NewRowIndex).

sub_matrix_rec([],_,_).
sub_matrix_rec([Element|Row],RowIndex,ColumnIndex):-
  get_element(RowIndex,ColumnIndex,IJ),
  NewElement is IJ-Element,
  replace_element(RowIndex,ColumnIndex,NewElement),!,
  NewColumnIndex is ColumnIndex+1,
  sub_matrix_rec(Row,RowIndex,NewColumnIndex).

%! transpose_matrix(+Matrix, -Transposed)
%
% Transpose Matrix and store it in the global variable (which is now the transposed
% matrix)

transpose_matrix(Matrix,TransposedMatrix):-
  transpose_matrix(Matrix),
  nb_getval(stored_matrix,TransposedMatrix).

transpose_matrix([Head|Tail]):-
  length([Head|Tail],ColumnSize), %These are column and row size of the TRANSPOSED matrix
  length(Head,RowSize),
  constant_matrix(RowSize,ColumnSize,0,Matrix),!,
  nb_setval(stored_matrix,Matrix),
  nb_setval(formal_m,RowSize),
  nb_setval(formal_n,ColumnSize),
  fill_matrix_transposed([Head|Tail],1).

fill_matrix_transposed([],_).
fill_matrix_transposed([Head|Tail],RowIndex):-
  fill_matrix_transposed_rec(Head,RowIndex,1),
  NewRowIndex is RowIndex+1,
  fill_matrix_transposed(Tail,NewRowIndex).

fill_matrix_transposed_rec([],_,_).
fill_matrix_transposed_rec([Head|Tail],RowIndex,ColumnIndex):-
  replace_element(ColumnIndex,RowIndex,Head),!,
  NewColumnIndex is ColumnIndex+1,
  fill_matrix_transposed_rec(Tail,RowIndex,NewColumnIndex).

%! lusubst(+B, -X)
%
% Use LU decomposition to solve Ax=b by fwd/backwd substitution
% Supposing A has already been decomposed using the ludcmp predicate

lusubst(B,X):-
   nb_getval(formal_n,N),
   nb_setval(formal_vector,B),
   nb_setval(formal_sum,0),
   forall(
         between(1,N,I),
         (nb_getval(formal_index, Index), nth1(I,Index,IP),
         nb_getval(formal_vector,Vector), nth1(IP,Vector,Y),
         nb_setval(formal_sum,Y),
         nth1(I,Vector,TMP),
         replace(Vector,IP,TMP,NewVector),nb_setval(formal_vector,NewVector),
         II is I-1,
         forall(
               between(1,II,J),
               (nb_getval(formal_sum,S),
               get_element(I,J,A),
               nth1(J,NewVector,YY),
               SS is S - A*YY,
               nb_setval(formal_sum,SS)
               )
         ),
         nb_getval(formal_sum,S),
         replace(NewVector,I,S,FinalNewVector),nb_setval(formal_vector,FinalNewVector)
         )
   ),
   nb_getval(formal_vector,VVector),nth1(N,VVector,Z),
   get_element(N,N,BB),
   ZZ is Z/BB,
   replace(VVector,N,ZZ,NewVVector), nb_setval(formal_vector,NewVVector),
   forall(
         between(2,N,I),
         (nb_setval(formal_sum,0),
         II is N-I+1,
         III is II+1,
         forall(
               between(III,N,J),
               (nb_getval(formal_sum,S),
               get_element(II,J,A),
               nb_getval(formal_vector,Vector),nth1(J,Vector,Y),
               SS is S + A*Y,
               nb_setval(formal_sum,SS)
               )
         ),
         nb_getval(formal_vector,Vector),nth1(II,Vector,Y),
         nb_getval(formal_sum,S),
         get_element(II,II,BBB),
         YY is (Y - S)/BBB,
         replace(Vector,II,YY,NewVector),nb_setval(formal_vector,NewVector)
         )
   ),
   nb_getval(formal_vector,X).


%! solve_linear_problem(+A, +B, -X)
%
% Solve the pbm A.X = B through delegate to ludcmp and lusubst
 
solve_linear_problem(A,B,X) :-
  ludcmp_matrix(A, _Decomposed_A),
  lusubst(B,X).
