/*
Defined the internal structure of biocham used to adress the models and items.
Most object of biocham are described as items through the item\4 dynamic predicate of
the form item(Id, Parent, Kind, Item), where:
Id: is a unique identifier generated by biocham
Parent: is the upper structure to which an object is attached, generaly a model
Kind: is a type-like information about the object, it may be: model, reaction, etc.
Item: the actual description of the object

Items may also have Keys, tag-like information that are handled separately through the
key\2 dynamic predicate.

Coder: T. Martinez
*/

:- module(
  models,
  [
    % Grammars
    range/1,
    ref/1,
    % Commands
    load/1,
    add/1,
    load_biocham/1,
    add_biocham/1,
    export_biocham/1,
    new_model/0,
    new_model/1,
    clear_model/0,
    list_models/0,
    list_current_models/0,
    list_model/0,
    select_model/1,
    set_model_name/1,
    inherits/1,
    parametrize/0,
    check_model/0,
    % Public API
    get_model_name/1,
    get_model_name/2,
    current_models/1,
    single_model/1,
    new_model/2,
    set_current_models/1,
    add_item/1,
    replace_item/4,
    change_item/4,
    set_annotation/3,
    get_annotation/3,
    delete_annotation/2,
    item/1,
    find_item/1,
    all_items/2,
    all_ids/2,
    list_items/1,
    list_ids/1,
    list_ids/2,
    print_item/1,
    deleting/1,
    delete_item/1,
    delete_items/1,
    add_dependency/2,
    add_file_suffix/2,
    inherits/2,
    inherits_from/2,
    directly_inherits_from/2,
    begin_command/0,
    get_selection/3,
    set_selection/3,
    at_delete/2,
    get_parent/2,
    get_model/2,
    load_all/2,
    parametrize_initial_concentration/0,
    parametrize_reaction/0,
    test_wf_reactant/2,
    test_wf_inhibitor/2,
    test_wf_other/2,
    test_strictness/2
  ]
).

% Only for separate compilation/linting
:- use_module(arithmetic_rules).
:- use_module(doc).
:- use_module(formal_derivation).
:- use_module(kinetics).
:- use_module(reaction_rules).
:- use_module(util).

:- devdoc('This secton needs documentation for the developper, in particular about the invariants concerning item, parent, id, key, kind. 

The error unkown_item we get is also unsufficiently informative.').

:- devdoc('Basically, you can change item by "change_item([parent: Id], initial_concentration, Variable, init(Variable = Value))" and access item by "item([parent: Id, kind: initial_concentration, key: Variable, item: init(Variable = Value)])" or find_item which raises an error if item fails.').

:-devdoc('
Warning: you cannot delete items through backtracking on the item predicate, you must use all_items and backtrack with member to delete items in the list.

Par exemple pour le retrait/ajout de règles dans revision.pl, j ai dû
directement manipuler les items de type (kind) reaction.

L API de models.pl donne quelques facilités, par exemple
 findall(Item, item([kind: reaction, item: Item]), Reactions)
s écrit plus simplement
 all_items([kind: reaction], Reactions)

et tu peux plus efficacement manipuler les ids
 all_ids([kind: reaction], Ids)
de façon à pouvoir ensuite appeler directement delete_item(Id)
plutôt que de devoir rechercher les réactions dans la base.
').

         
:- devdoc('\\section{Grammars}').


:- grammar(range).


range(Integer0 - Integer1) :-
  integer(Integer0),
  integer(Integer1).

range(Integer) :-
  integer(Integer).


:- grammar(ref).


ref(Range) :-
  list(range, Range).

ref(Name) :-
  name(Name).


:- devdoc('\\section{Commands}').


load(InputFile) :-
  biocham_command,
  type(InputFile, input_file),
  doc('
    acts as the corresponding \\command{load_biocham/1} / \\command{load_sbml/1}
    / \\command{import_ode/1} / \\command{load_table/1},
    depending on the file extension
    (respectively \\texttt{.bc}, \\texttt{.xml}, \\texttt{.ode}, \\texttt{.csv}
    -- assuming no extension is \\texttt{.bc}).'),
  load_all(_Suffix, InputFile).


add(InputFile) :-
  biocham_command,
  type(InputFile, input_file),
  doc('
    acts as the corresponding \\command{add_biocham/1} / \\command{add_sbml/1}
    / \\command{add_ode/1} / \\command{add_table/1},
    depending on the file extension
    (respectively \\texttt{.bc}, \\texttt{.xml}, \\texttt{.ode}, \\texttt{.csv}
    -- assuming no extension is \\texttt{.bc}).'),
  add_all(_Suffix, InputFile).


load_biocham(InputFile) :-
  biocham_command,
  type(InputFile, input_file),
  doc('
    opens a new model, loads the reaction rules and executes the commands
    (with the file directory as current directory)
    contained in the given Biocham \\texttt{.bc} file.
    The suffix \\texttt{.bc} is automatically added to the name if such a
    file exists.'),
  load_all('bc', InputFile).


add_biocham(InputFile) :-
  biocham_command,
  type(InputFile, input_file),
  doc('
    the rules of the given \\texttt{.bc} file are loaded and
    \\emph{added} to the current set of rules.
    The commands contained in the file are executed
    (with the file directory as current directory).'),
  add_all('bc', InputFile).


export_biocham(OutputFile) :-
  biocham_command,
  type(OutputFile, output_file),
  doc('
    exports the current model into a \\texttt{.bc} file.
  '),
  with_output_to_file(OutputFile, list_model).


new_model :-
  biocham_command,
  doc('opens a new fresh model.'),
  (
    fresh,
    current_models([_Id])
  ->
    true
  ;
    new_model(Id, _ModelName),
    set_current_models(Id),
    assertz(fresh)
  ).

new_model(ModelName) :-
  biocham_command,
  doc('opens a new fresh model with a selected name.'),
  (
    fresh,
    current_models([_Id]),
    set_model_name(ModelName)
  ->
    true
  ;
    new_model(Id, ModelName),
    set_current_models(Id),
    assertz(fresh)
  ).


clear_model :-
  biocham_command,
  doc('clears the current model.'),
  current_models(CurrentModels),
  forall(
    member(Id, CurrentModels)
  ,
    delete_item(Id)
  ),
  set_option(show, {}),
  retractall(nusmv:query_molecules(_)),
  set_counter(molecule_id, 0),
  set_counter(parameter_id, 1),
  assertz(nusmv:query_molecules([])).


list_models :-
  biocham_command,
  doc('lists all open models.'),
  list_items([parent: top, kind: model]).


list_current_models :-
  biocham_command,
  doc('lists current models.'),
  current_models(CurrentModels),
  list_ids(CurrentModels).


list_model :-
  biocham_command,
  doc('lists the current Biocham model: reactions, influences, initial state, parameters, events, functions, LTL patterns, options.'),
  list_model_reactions,
  list_model_influences,
  list_model_initial_state,
  list_model_parameters,
  list_model_events,
  list_model_macros(function),
  list_model_macros(ltl_pattern),
  list_model_options.


select_model(RefSet) :-
  biocham_command,
  type(RefSet, {ref}),
  doc('selects some models.'),
  find_model_refs(RefSet, CurrentModels),
  set_current_models(CurrentModels).


set_model_name(Name) :-
  biocham_command,
  type(Name, name),
  doc('changes the current model name.'),
  single_model(Id),
  replace_item(Id, model, Name, Name).


inherits(Ancestors) :-
  biocham_command(*),
  type(Ancestors, '*'(ref)),
  doc('makes the current model inherit from the given ancestor models.'),
  single_model(Id),
  find_model_refs(Ancestors, Models),
  forall(
    member(AncestorId, Models)
  ,
    inherits(Id, AncestorId)
  ).

%! parametrize
%
% Replace numeric constants in a model by parameters initialized with their value

parametrize :-
  biocham_command,
  doc('Replace numeric constants in a model by parameters k1, k2,... initialized with their value.'),
  parametrize_initial_concentration,
  parametrize_reaction.


check_model :-
  biocham_command,
  doc('Check if the current model is well formed and strict, (see TCS15_FGS).'),
  doc('If it is not the case, proposes a witness.'),
  doc('Check if the current ODE system is polynomial.'),
  single_model(ModelId),
  findall(
    Molecule,
    identifier_kind(ModelId, Molecule, object),
    List_Molecule
  ),
  (
    List_Molecule = []
  ->
    format("There is currently no model to work on!~n", [])
  ;
    reactions_with_species(List_Molecule, List_Reaction),
    (
      forall(
        member(Reaction, List_Reaction),
        (
          test_wellformed(List_Molecule, Reaction)
        ;
          format("Due to: ~w ; ", [Reaction]),
          fail
        )
      )
    ->
      format("The model is wellformed.~n", [])
    ;
      format("the model is NOT wellformed.~n", [])
    ),
    (
      test_strictness(List_Molecule, List_Reaction)
    ->
      format("The model is strict.~n", [])
    ;
      format("The model is NOT strict.~n", [])
    )
  ),
  (
    with_current_ode_system(
      forall(
        ode(_Var, Expr),
        is_polynomial(Expr)
      )
    )
  ->
    format('The current ode system is polynomial.~n', [])
  ;
    format('The current ode system is NOT polynomial.~n', [])
  ).


:- devdoc('\\section{Public API}').



%! get_model_name((+Id), -Name)
%
% Return the name of the current (or pointed) model

get_model_name(Name) :-
  single_model(Id),
  get_model_name(Id, Name).

get_model_name(Id, Name) :-
  find_item([parent: top, id: Id, item: Name]).


current_models(Models) :-
  get_selection(top, current_models, Models).


%! single_model(?Model)
%
% Check that the current selection is a single model (and return it)

single_model(Model) :-
  (
    current_models([Model])
  ->
    true
  ;
    throw(error(need_single_model))
  ).


%! new_model(-Id, ?ModelName)
%
% Create a new model and return its Id and name(, name may be prescribed). 

new_model(Id, ModelName) :-
  ( %default model name to new_model if still a variable
    var(ModelName)
  ->
    count(model, N),
    atomic_concat("model_", N, ModelName)
  ;
    true
  ),
  add_item([parent: top, kind: model, key: ModelName, id: Id]),
  (
    item([parent: top, kind: model, key: initial, id: InitialId])
  ->
    inherits(Id, InitialId)
  ;
    true
  ).


%! set_current_models(+Models)
%
% Set as selected the list of models

set_current_models(Models) :-
  (
    Models = []
  ->
    throw(error(cannot_select_no_models))
  ;
    \+(is_list(Models)) % In case of a single model is desired
  ->
    set_selection(top, current_models, [Models])
  ;
    set_selection(top, current_models, Models)
  ).


%! add_item(?Options)
%
% front-end dealing with options and error for put_item
% both rely on the dynamic item\4 predicate

add_item(Options) :-
  memberchk(kind: Kind, Options),
  optional(item: Item, Options),
  optional(parent: Parent, Options),
  optional(key: Key, Options),
  optional(id: Id, Options),
  default(Parent, current_model),
  default(Key, []),
  (
    Parent = current_model
  ->
    single_model(Parent0)
  ;
    Parent0 = Parent
  ),
  (
    var(Item)
  ->
    (
      list(Key)
    ->
      throw(error(no_item_given))
    ;
      Item = Key
    )
  ;
    true
  ),
  not_fresh,
  count(item_id, Id),
  put_item(Parent0, Kind, Key, Item, Id).


replace_item(Id, Kind, Key, Item) :-
  retract(item(Id, Parent, _Kind, _Item)),
  retractall(key(_Key, Id)),
  put_item(Parent, Kind, Key, Item, Id).


change_item(Options, Kind, Key, Item) :-
  catch(
    (
      find_item([noinheritance, kind: Kind, key: Key, id: Id | Options]),
      replace_item(Id, Kind, Key, Item)
    ),
    error(unknown_item(_)),
    add_item([kind: Kind, key: Key, item: Item | Options])
  ).


%! item(+Options)
%
% used to retrieve information about an item
% for example item([id:137, kind:Kind]). allow to fetch the kind of item 137

item(Options) :-
  optional(kind: Kind, Options),
  optional(id: Id, Options),
  optional(key: Key, Options),
  optional(item: Item, Options),
  item_parent_option(Options, Parent),
  (
    var(Id)
  ->
    (
      Parent0 = Parent
    ;
      \+ member(no_inheritance, Options),
      inherits_from(Parent, Parent0)
    )
  ;
    Parent0 = Parent
  ),
  (
    list(Kind)
  ->
    member(Kind0, Kind)
  ;
    Kind0 = Kind
  ),
  (
    var(Key)
  ->
    true
  ;
    list(Key)
  ->
    indexes_ids(Key, Ids),
    member(Id, Ids)
  ;
    key(Key, Id)
  ),
  item(Id, Parent0, Kind0, Item).


find_item(Options) :-
  (
    item(Options)
  ->
    true
  ;
    throw(error(unknown_item(Options)))
  ).


all_items(Options, Items) :-
  findall(
    Item,
    item([item: Item | Options]),
    Items
  ).


all_ids(Options, Ids) :-
  findall(
    Id,
    item([id: Id | Options]),
    Ids
  ).


list_items(Options) :-
  all_ids([no_inheritance | Options], Ids),
  list_ids(Options, Ids),
  (
    member(no_inheritance, Options)
  ->
    true
  ;
    item_parent_option(Options, Parent),
    (
      var(Parent)
    ->
      true
    ;
      optional(indent: Indent, Options),
      default(Indent, 0),
      forall(
        (inherits_from(Parent, Parent0),
        all_ids([parent: Parent0, no_inheritance | Options], InheritedIds),
        InheritedIds \= [])
      ,
        (indent(Indent),
        find_item([id: Parent0, item: ParentName]),
        format('From inherited \'~w\':\n', [ParentName]),
        list_ids(Options, InheritedIds))
      )
    )
  ).


list_ids(Ids) :-
  list_ids([], Ids).


list_ids(Options, Ids) :-
  (
    peek_count(list_item_counter, 0),
    \+ memberchk(no_numbering, Options)
  ->
    retractall(listed_item(_, _))
  ;
    true
  ),
  list_ids_aux(Options, Ids).


print_item(Id) :-
  item(Id, _Parent, Kind, Item),
  (
    atom_concat('print_', Kind, F),
    G =.. [F, Id, Item],
    predicate_property(G, visible)
  ->
    G
  ;
    print(Item)
  ).


% make it so that print/1 does not quote things that look like a variable
user:portray(Thing) :-
  Thing =.. [Functor | Args],
  catch(
    read_term_from_atom(Functor, Term, []),
    _,
    fail
  ),
  var(Term),
  !,
  maplist(print_to_atom, Args, PArgs),
  atomic_list_concat(PArgs, ', ', AArgs),
  (
    Args == []
  ->
    write(Functor)
  ;
    format("~w(~w)", [Functor, AArgs])
  ).


print_to_atom(Term, Atom) :-
  format(atom(Atom), '~p', [Term]).


:- dynamic(deleting/1).


delete_item(Options) :-
  list(Options),
  !,
  find_item([id: Id | Options]),
  delete_item(Id).


delete_item(Id) :-
  get_parent(Id, Parent),
  delete_item_aux(Id),
  update_parent_model_identifier_kinds(Parent).


delete_item_aux(Id) :-
  (
    deleting(Id)
 ->
    throw(error(circular_delete(Id), delete_item_aux))
  ;
    assertz(deleting(Id))
  ),
  forall(
    at_delete_goal(Id, Goal)
  ,
    (
      Goal
    ->
      true
    ;
      throw(error(delete_goal_failure(Id), delete_item_aux))
    )
  ),
  retract(item(Id, _Model, KindItem, _Item)),
  retractall(key(_Key, Id)),
  retractall(annotation(Id, _KindAnnotation, _Annotation)),
  retractall(listed_item(_Index, Id)),
  retractall(dependency(Id, _Master)),
  retractall(at_delete_goal(Id, _Goal)),
  forall(
    item(SubId, Id, _SubKind, _SubItem)
  ,
    (
      item(SubId, __Id, __SubKind, __SubItem)
    ->
      delete_item_aux(SubId)
    ;
      true
    )
  ),
  forall(
    dependency(SubId, Id)
  ,
    (
      item(SubId, ___Id, ___SubKind, ___SubItem)
    ->
      delete_item_aux(SubId)
    ;
      true
    )
  ),
  retractall(selection(_, _, Id)),
  (
    KindItem = model
  ->
    (
      selection(top, current_models, _)
    ->
      true
    ;
      new_model
    )
  ;
    true
  ),
  retract(deleting(Id)).


delete_items(Options) :-
  forall(
    item([id: Id | Options])
  ,
    delete_item(Id)
  ).


%%% ANNOTATIONS %%%

%! set_annotation(+Id, +Kind, +Annotation)

set_annotation(Id, Kind, Annotation) :-
  retractall(annotation(Id, Kind, _)),
  assertz(annotation(Id, Kind, Annotation)).


%! get_annotation(+Id, +Kind, -Annotation)

get_annotation(Id, Kind, Annotation) :-
  annotation(Id, Kind, Annotation).


delete_annotation(Id, Kind) :-
  retract(annotation(Id, Kind, _)).

add_dependency(IdSub, IdMaster) :-
  assertz(dependency(IdSub, IdMaster)).


:- multifile(add_file_suffix/2).


add_file_suffix('', add_biocham_file).


add_file_suffix('bc', add_biocham_file).


%%% HERITANCE %%%

%! inherits(+Id, +Ancestor_id)
%
% Safe constructor for the directly_inherits_from\2 predicate

inherits(Id, AncestorId) :-
  (
    Id = AncestorId
  ->
    throw(error(cannot_inherit_from_itself, inherits))
  ;
    self_or_inherits_from(Child, Id),
    self_or_inherits_from(AncestorId, Parent),
    inherits_from(Parent, Child)
  ->
    throw(error(cannot_inherit_from_descendant, inherits))
  ;
    (
      item(Id, top, model, _Item0),
      item(AncestorId, top, model, _Item1),
      self_or_inherits_from(Child, Id),
      identifier_kind(Child, Ident, Kind0),
      Kind0 \= free,
      identifier_kind(AncestorId, Ident, Kind1),
      Kind1 \= free,
      Kind0 \= Kind1
    ->
      throw(error(kind_mismatch(Ident, Kind0, Kind1), inherits))
    ;
      true
    ),
    assertz(directly_inherits_from(Id, AncestorId))
  ).


prolog:message(error(cannot_inherit_from_itself)) -->
  ['Cannot inherit from itself.'].


prolog:message(error(already_inherits)) -->
  ['There is already such an inheritance relation.'].


prolog:message(error(cannot_inherit_from_descendant)) -->
  ['Cannot inherit from descendant.'].

prolog:message(error(kind_mismatch(Ident, Kind, NeededKind))) -->
  ['There is a mismatch between the uses of ~a: ~a and ~a.'-[Ident,
    Kind, NeededKind]].

%! inherits_from(?Id, ?AncestorId)
%
% recursive check of heritance

inherits_from(Id, AncestorId) :-
  directly_inherits_from(Id, AncestorId).

inherits_from(Id, AncestorId) :-
  directly_inherits_from(Id, IntermediateAncestorId),
  inherits_from(IntermediateAncestorId, AncestorId).


self_or_inherits_from(Id, Id).

self_or_inherits_from(Id, AncestorId) :-
    inherits_from(Id, AncestorId).


:- dynamic(directly_inherits_from/2).


begin_command :-
  set_counter(list_item_counter, 0).


%! get_selection(+ModelName, +SelectionName, -Ids)
%
% Mainly used to retrieve the id of current models

get_selection(ParentOrCurrentModel, SelectionName, Ids) :-
  get_selection_parent(ParentOrCurrentModel, Parent),
  findall(
    Id,
    selection(Parent, SelectionName, Id),
    Ids
  ).


%! set_selection(+ModelName, +SelectionName, -SpecifierList)
%
% Create a selection fo all specifier, they may be option like or ids

set_selection(ParentOrCurrentModel, SelectionName, IdsOrOptions) :-
  get_selection_parent(ParentOrCurrentModel, Parent),
  retractall(selection(Parent, SelectionName, _)),
  forall(
    (
      IdsOrOptions = [_: _| _]
    ->
      item([id: Id | IdsOrOptions])
    ;
      member(Id, IdsOrOptions)
    )
  ,
    assertz(selection(Parent, SelectionName, Id))
  ).


at_delete(Id, Goal) :-
  assertz(at_delete_goal(Id, Goal)).


get_parent(Id, ParentId) :-
  item(Id, ParentId, _, _).


%! get_model(+Id, -ModelId)
%
% Determine the id of the parent model from a given item

get_model(Id, ModelId) :-
  item(Id, ParentId, Kind, _),
  (
    Kind = model
  ->
    ModelId = Id
  ;
    get_model(ParentId, ModelId)
  ).


:- devdoc('\\section{Private predicates}').


optional(Item, List) :-
  (
    member(Item, List)
  ->
    true
  ;
    true
  ).


default(Variable, Value) :-
  (
    var(Variable)
  ->
    Variable = Value
  ;
    true
  ).


put_item(Parent, Kind, Key, Item, Id) :-
  (
    list(Key)
  ->
    forall(
      member(K, Key)
    ,
      assertz(key(K, Id))
    )
  ;
    assertz(key(Key, Id))
  ),
  assertz(item(Id, Parent, Kind, Item)),
  update_parent_model_identifier_kinds(Parent).


item_parent_option(Options, ChosenParent) :-
  optional(parent: Parent, Options),
  optional(id: Id, Options),
  (
    var(Id)
  ->
    default(Parent, current_model),
    (
      Parent == current_model
    ->
      single_model(ChosenParent)
    ;
      Parent == current_models
    ->
      current_models(Models),
      member(ChosenParent, Models)
    ;
      ChosenParent = Parent
    )
  ;
    ChosenParent = Parent
  ).


:- dynamic(listed_item/2).


list_ids_aux(Options, Ids) :-
  optional(indent: Indent, Options),
  default(Indent, 0),
  (
    member(recursive, Options)
  ->
    Recursive = yes
  ;
    Recursive = no
  ),
  forall(
    member(Id, Ids)
  ,
    (
      indent(Indent),
      (
        selection(_, _, Id)
      ->
        Selected = '*'
      ;
        Selected = ' '
      ),
      (
        memberchk(no_numbering, Options)
      ->
        true
      ;
        count(list_item_counter, Counter),
        assertz(listed_item(Counter, Id)),
        format('[~d]', [Counter])
      ),
      write(Selected),
      print_item(Id),
      nl,
      (
        Recursive = yes
      ->
        SubIndent is Indent + 1,
        all_ids([parent: Id | Options], SubIds),
        list_ids_aux([indent: SubIndent, recursive | Options], SubIds)
      ;
        true
      )
    )
  ).


indexes_ids(Indexes, Ids) :-
  findall(
    Id,
    (
      member(Ranges, Indexes),
      ranges_ids(Ranges, RangeIds),
      member(Id, RangeIds)
    ),
    Ids
  ).


ranges_ids(Ranges, Ids) :-
  findall(
    Id,
    (
      member(Range, Ranges),
      (
        Range = Min - Max
      ->
        (
          Min =< Max
        ->
          between(Min, Max, Range)
        ;
          throw(error(invalid_range(Min, Max)))
        )
      ;
        Index = Range
      ),
      resolve_index(Index, Id)
    ),
    Ids
  ).


resolve_index(Index, Id) :-
  (
    listed_item(Index, Id)
  ->
    true
  ;
    throw(error(no_such_element, Index))
  ).




:- dynamic(dependency/2).


:- dynamic(fresh/0).


:- dynamic(at_delete_goal/2).


:- dynamic(selection/3).


get_selection_parent(ParentOrCurrentModel, Parent) :-
  (
    ParentOrCurrentModel = current_model
  ->
    single_model(Parent)
  ;
    Parent = ParentOrCurrentModel
  ).


load_all(Suffix, InputFile) :-
  current_models(OldCurrentModels),
  findall(
    Id,
    (
      filename(InputFile, Filename),
      (
        var(Suffix)
      ->
        file_name_extension(_, Suffix, Filename)
      ;
        true
      ),
      load(Suffix, Filename),
      current_models(Ids),
      member(Id, Ids)
    ),
    NewCurrentModels
  ),
  (
    NewCurrentModels = []
  ->
    restore_current_models(OldCurrentModels)
  ;
    set_current_models(NewCurrentModels)
  ).


restore_current_models(OldCurrentModels) :-
  findall(
    Model,
    (
      member(Model, OldCurrentModels),
      once(item([id: Model]))
    ),
    CurrentModels
  ),
  (
    CurrentModels = []
  ->
    new_model
  ;
    set_current_models(CurrentModels)
  ).


add_all(Suffix, InputFile) :-
  forall(
    filename(InputFile, Filename)
  ,
    (
      (
        var(Suffix)
      ->
        file_name_extension(_, Suffix, Filename)
      ;
        true
      ),
      add(Suffix, Filename)
    )
  ).


load(Suffix, Filename) :-
  new_model,
  file_base_name(Filename, FileBaseName),
  (
    file_name_extension(Base, _, FileBaseName)
  ->
    true
  ;
    Base = FileBaseName
  ),
  set_model_name(Base),
  add(Suffix, Filename).


add(Suffix, Filename) :- % e.g. calls add_ode_system_file(Filename) with add_suffix defined in ode.pl
  (
    add_file_suffix(Suffix, Functor)
  ->
    call(Functor, Filename)
  ;
    throw(error(unknown_suffix(Suffix)))
  ).


add_biocham_file(Filename) :-
  automatic_suffix(Filename, '.bc', read, FilenameBc),
  file_directory_name(FilenameBc, FileDirectory),
  setup_call_cleanup(
    (
      open(FilenameBc, read, Stream),
      working_directory(PreviousDirectory, FileDirectory)
    ),
    load_biocham_stream(Stream),
    (
      close(Stream),
      working_directory(_, PreviousDirectory)
    )
  ).


load_biocham_stream(Stream) :-
  \+ (
    repeat,
    read_term(Stream, Command, [variable_names(VariableNames)]),
    (
      Command = end_of_file
    ->
      !
    ;
      name_variables_and_anonymous(Command, VariableNames),
      command(Command)
    ),
    fail
  ).


not_fresh :-
  retractall(fresh).


:- dynamic(item/4).


:- dynamic(key/2).


:- dynamic(annotation/3).


update_parent_model_identifier_kinds(Parent) :-
  (
    Parent = top
  ->
    true
  ;
    get_model(Parent, ModelId),
    update_identifier_kinds(ModelId)
  ).


find_model_refs(RefSet, Models) :-
  findall(
    Id,
    (
      member(Ref, RefSet),
      (
        list(Ref)
      ->
        ranges_ids(Ref, Ids),
        member(Id, Ids)
      ;
        find_item([parent: top, key: Ref, id: Id])
      ),
      item([id: Id, kind: Kind]),
      (
        Kind = model
      ->
        true
      ;
        throw(error(not_a_model(Kind)))
      )
    ),
    Models
  ).


%! parametrize_initial_concentration 
%
% replace numeral initial concentration by parameter with the same value

parametrize_initial_concentration :-
  forall(
    (
      item([item: present(Species, Value), kind:initial_state])
    ;
      (item([item: present(Species), kind:initial_state]) ,Value=1)
    )
  ,
    (
      number(Value)
    ->
      new_parameter(K),
      command(present(Species, K)),
      command(parameter(K = Value))
    ;
      true
    )
  ).


%! parametrize_reaction
%
% Replace the numeral of a kinetic by similar parameter

parametrize_reaction :-
  forall(
    item([item: Reac, kind: reaction, id: Id])
  ,
    (delete_item(Id),
    introduce_param(Reac, Reac2),
    command(Reac2))
  ).


%! introduce_param(+reaction, -parametrized_reaction)
%
% recursive routine for parametrize_reaction

introduce_param(Kin for Reac, Kin2 for Reac) :-
  introduce_param(Kin, Kin2).

introduce_param(A*B, A2*B2) :-
  introduce_param(A, A2),
  introduce_param(B, B2).

introduce_param(A/B, A2/B2) :-
  introduce_param(A, A2),
  introduce_param(B, B2).

introduce_param(A+B, A2+B2) :-
  introduce_param(A, A2),
  introduce_param(B, B2).

introduce_param(A-B, A2-B2) :-
  introduce_param(A, A2),
  introduce_param(B, B2).

introduce_param(A^N, A2^N) :-
  introduce_param(A, A2).

introduce_param(X, K) :-
  number(X),
  new_parameter(K),
  command(parameter(K = X)).

introduce_param(Func, Func2) :-
  Func =.. [Name | Param],
  maplist(introduce_param, Param, Param2),
  Func2 =.. [Name | Param2].

introduce_param(O,O).


%! formal_reactant(+Molecule, +Reaction)

formal_reactant(Mol, Reactants/_Inhibtors => _Products) :- !,
  is_present(Mol, Reactants).

formal_reactant(Mol, Reactants => _Products) :- !,
  is_present(Mol, Reactants).


%! formal_product(+Molecule, +Reaction)

formal_product(Mol, _Reactants/_Inhibtors => Products) :- !,
  is_present(Mol, Products).

formal_product(Mol, _Reactants => Products) :- !,
  is_present(Mol, Products).


%! formal_catalyst(+Molecule, +Reaction)

formal_catalyst(Mol, Reactants => Products) :- !,
  is_present(Mol, Reactants),
  is_present(Mol, Products).

formal_catalyst(Mol, Reactants/_Inhibtors => Products) :- !,
  is_present(Mol, Reactants),
  is_present(Mol, Products).

%! formal_inhibitor(+Molecule, +Reaction)

formal_inhibitor(Mol, _Reactants/Inhibitors => _Products) :- !,
  is_present(Mol, Inhibitors).


% is_present(Mol, Expr)

% is_present(A, A) :- !.
% is_present(A, _K*A) :- !.
% is_present(A, _Tail+A) :- !.
% is_present(A, _Tail+_K*A) :- !.
% is_present(A, Tail+_H) :- is_present(A, Tail).


%! test_wellformed(+List_Molecule, +Reaction)
%
% loop on molecule to determine if they follow the condition of wellformedness

test_wellformed([], _Reaction).

test_wellformed(_Molecules, 'MA'(_R) for _Reaction).

test_wellformed([Molecule|Tail], Expression_raw for Reaction) :-
  ode:substitute_functions(Expression_raw, Expression),
  always_positive(Expression),
  (
    % detect null kinetics as a warning
    is_null(Expression)
  ->
    format("WARNING: Some kinetics are detected as null: ~w~n", [Expression]),
    fail
  ;
    formal_reactant(Molecule, Reaction)
  ->
    present_in_kinetics(Molecule, Expression),
    test_wf_reactant(Molecule, Expression)
  ;
    formal_inhibitor(Molecule, Reaction)
  ->
    present_in_kinetics(Molecule, Expression),
    test_wf_inhibitor(Molecule, Expression)
  ;
    !,
    (
      \+(present_in_kinetics(Molecule, Expression)),
      !
    ;
      test_wf_other(Molecule, Expression)
    )
  ),
  test_wellformed(Tail, Expression for Reaction).


%! test_wf_reactant(+Molecule, +Expression)
%
% Check if the derivative of Expression wrt Molecule is positive somewhere

test_wf_reactant(Molecule, Expression) :-
  formal_derivation:derivate(Expression, Molecule, Derivative),
  (
    \+(arithmetic_rules:always_negative(Derivative))
  ;
    format("Fail for wellformedness of ~w with ~w as reactant.~n", [Expression, Molecule]),
    fail
  ),!.

%! test_wf_inhibitor(+Molecule, +Expression)
%
% Check if the derivative of Expression wrt Molecule is negative somewhere

test_wf_inhibitor(Molecule, Expression) :-
  formal_derivation:derivate(Expression, Molecule, Derivative),
  (
    \+(arithmetic_rules:always_positive(Derivative))
  ;
    format("Fail for wellformedness of ~w with ~w as inhibitor.~n", [Expression, Molecule]),
    fail
  ),!.

%! test_wf_other(+Molecule, +Expression)

test_wf_other(Molecule, Expression) :-
  formal_derivation:derivate(Expression, Molecule, Derivative),
  (
    is_null(Derivative)
  ;
    format("Fail for wellformedness of ~w with ~w as not present.~n", [Expression, Molecule]),
    fail
  ),!.


%! test_strictness(+List_Molecule, +List_Reaction)
%
% Check if Expression is null when Molecule is absent

test_strictness(_List_Molecules, []).

test_strictness(List_Molecules, [Expr for Reaction|TailReac]) :-
  (
    Expr = 'MA'(_R)
  ->
    true
  ;
    ode:substitute_functions(Expr, Expression),
    test_strictness_sr(List_Molecules, Expression for Reaction)
  ),
  test_strictness(List_Molecules, TailReac).


test_strictness_sr([], _).

test_strictness_sr([Molecule|Tail], Expression for Reaction) :-
  formal_reactant(Molecule, Reaction),
  !,
  substitute([Molecule], [0], Expression, Output),
  (
    is_null(Output)
  ->
    test_strictness_sr(Tail, Expression for Reaction)
  ;
    format("Due to: ~w while checking ~w;~n", [Expression, Molecule]),
    fail
  ).

test_strictness_sr([_Molecule|Tail], Expression for Reaction) :-
  test_strictness_sr(Tail, Expression for Reaction).

