/** <PODE> Tools to transform an ODE into a PODE
 * 
 * By definition in this module, the free variable is supposed to be the time and denoted
 * as 't'.
 *
@author M. Hemery
@license GNU-GPL v.2
*/

:- module(
  pode,
  [
    %Commands
    polynomize_ode/0,
    %API
    polynomize_ode/3,
    complete_derivation/3
  ]).

:- use_module(biocham).
:- use_module(counters).
:- use_module(doc).
:- use_module(formal_derivation).
:- use_module(objects).
:- use_module(reaction_rules).
:- use_module(util).

% BIOCHAM Command

polynomize_ode :-
  biocham_command,
  doc('Fetch the current system of ODE and apply the Bournez algorithm: this introduces new variables into the system until the whole ODE is polynomial.'),
  % Fetch the old system (TODO: handle the parameters, functions and co.)
  with_current_ode_system((
    get_current_ode_system(Id),
    findall(Deriv, item([parent:Id, kind:ode, item:Deriv]), Input)
  )),
  % Determine the transformation and the new system
  polynomize_ode(Input, Output, Transf_raw),
  sort(Transf_raw, Transf),
  forall(
    member(Name->Expr, Transf),
    format("~w = ~w~n", [Name, Expr])
  ),
  % Construct and select the new system
  new_ode_system(NewId),
  set_ode_system_name(NewId, polynomial_system),
  forall(
    member(Deriv, Output),
    add_item([parent: NewId, kind:ode, item:Deriv])
  ),
  select_ode_system(polynomial_system).


% API

%! polynomize_ode(+Input, -Output, -Transf) 
%
% Construct the list of variables and delegate to polynomize_ode\4

polynomize_ode(Input, Output, Transf) :-
  set_counter(variable_id, 0),
  maplist([D,V]>>(D=(d(V)/dt=_Deriv)), Input, List_of_variables),
  polynomize_ode(Input, [], [], List_of_variables, Transf, Output).


% Internal functions

%! polynomize_ode(+Input, + Accumulator, +In_Transf, +List_of_variables, -List_Transf, -Output) 
%
% 

polynomize_ode([], Output, Transf, _LoV, Transf, Output).

% IT/TT/OT => Input, Tempo, Output Transformation
polynomize_ode([d(Var)/dt = Deriv | Tail], Acc, IT, LoV, OT, Output) :-
  detect_new_variable(Deriv, LoV, List_raw),
  sort(List_raw, List), % remove duplicate
  (
    List = []
  ->
    polynomize_ode(Tail, [d(Var)/dt = Deriv | Acc], IT, LoV, OT, Output)
  ;
    add_variable(Acc, Var, Deriv, Tail, LoV, List, NewDeriv, NewTail, NewLoV, IT, TT),
    polynomize_ode(NewTail, [d(Var)/dt = NewDeriv | Acc], TT, NewLoV, OT, Output)
  ).


%! detect_new_variable(+Expr, +Current_variables, -List_of_expression)
%
% detect the expression for variables that needs to be introduced

detect_new_variable(E1+E2, LoV, L12) :-
  detect_new_variable(E1, LoV, L1),
  detect_new_variable(E2, LoV, L2),
  append(L1, L2, L12).

detect_new_variable(E1-E2, LoV, L12) :-
  detect_new_variable(E1, LoV, L1),
  detect_new_variable(E2, LoV, L2),
  append(L1, L2, L12).

detect_new_variable(-Expr, LoV, List) :-
  detect_new_variable(Expr, LoV, List).

detect_new_variable(E1*E2, LoV, L12) :-
  detect_new_variable(E1, LoV, L1),
  detect_new_variable(E2, LoV, L2),
  append(L1, L2, L12).

detect_new_variable(Expr^N, LoV, L) :-
  number(N),
  detect_new_variable(Expr, LoV, L).

detect_new_variable(Expr, _LoV, []) :-
  number(Expr),
  !.

detect_new_variable(Expr, LoV, []) :-
  member(Expr, LoV),
  !.

detect_new_variable(Expr, _LoV, [Expr]).


%! add_variable(+Acc_ODE, +Current_variable, +Current_derivative
%               +Reminder_ODE, +List_of_variable, +List_expression_to_add
%               -Modified_derivative, -Modified_reminder_ode, -Modified_list_variable
%               +Current_transformation, -Modified_transformation)
%
% add a variable to the current set of ODEs. Here transformation is the list of introduced 
% variables and their correspondance: [A->sin(x), etc.]

add_variable(_Acc, _Var, Deriv, Tail, LoV, [], Deriv, Tail, LoV, IT, IT) :- !.

add_variable(Acc, Var, Deriv, Tail, LoV, [Head|Other], NewDeriv, [d(Name)/dt = Deriv_Head|NewTail], [Name|LoVTempo], IT, OT) :-
  add_variable(Acc, Var, Deriv, Tail, LoV, Other, DerivTempo, TailTempo, LoVTempo, IT, TT),
  new_variable(LoV, Name),
  OT = [Name->Head|TT],
  maplist([A,B,C]>>(A=(B->C)), OT, Names, Heads), % []>>() is the lambda expr. syntax
  substitute(Heads, Names, DerivTempo, NewDeriv),
  substitute(Heads, Names, TailTempo, NewTail),
  append([Acc,[d(Var)/dt=NewDeriv],NewTail], List_Deriv),
  complete_derivation(Head, List_Deriv, Deriv_raw1),
  substitute(Heads, Names, Deriv_raw1, Deriv_raw2),
  simplify(Deriv_raw2, Deriv_Head).


%! complete_derivation(+Expr, +List_Deriv, -Deriv)
%
% Derivate Expr with respect to time and all variable in List_Deriv

complete_derivation(Expr, [], Timeder) :- !,
  derivate(Expr, t, Timeder).

complete_derivation(Expr, [d(Head)/dt = Der_ofH | Tail], Deriv + Der_wrtH*Der_ofH) :-
  derivate(Expr, Head, Der_wrtH), % derivate with respect to Head
  complete_derivation(Expr, Tail, Deriv).


%! new_variable(+LoV, -Name)
%
% Generate the name of a new (and not in LoV) variable of the form A,B,C,..,AA,BB, etc.

new_variable(LoV, Name) :-
  count(variable_id, N),
  divmod(N,26,R,C),
  Code is C+65,
  Rep is R+1,
  char_code(Letter, Code),
  util:stutter(Rep, Letter, Try),
  ( % test if the variable already exist
    member(Try, LoV)
  ->
    new_variable(LoV, Name)
  ;
    Name = Try
  ).
